@@ -1,5 +1,5 @@
 /*-
- * Copyright (c) 2008, 2009, 2012, 2013  Peter Pentchev
+ * Copyright (c) 2008, 2009, 2012, 2013, 2016  Peter Pentchev
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
@@ -22,15 +22,12 @@
  * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
  * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
  * SUCH DAMAGE.
- *
- * $Ringlet$
  */
 
-#define _GNU_SOURCE
-
 #include <ctype.h>
 #include <err.h>
 #include <fnmatch.h>
+#include <stdbool.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -44,6 +41,13 @@
 #include "confget_http_get.h"
 #include "confget_ini.h"
 
+struct var {
+	char	*name;
+	char	*value;
+	char	*section;
+	bool	 display;
+};
+
 /* The configuration filename specified by the user. */
 const char		*filename;
 /* The configuration section specified by the user. */
@@ -53,9 +57,11 @@ static int		 margc;
 /* The variables to extract or patterns to match against. */
 static char		* const *margv;
 /* Display the variable names, or only the values? */
-static int		 showvarname;
+static bool		 showvarname;
 /* Look for more variables after a match is found? */
-static int		 manyvars;
+static bool		 manyvars;
+/* Should section variables override default ones? */
+bool			 override;
 /* The pattern to match the variable value */
 static const char	*matchvalue;
 /* The prefix to use when displaying the variable name */
@@ -63,8 +69,8 @@ static const char	*output_prefix;
 /* The postfix to use when displaying the variable name */
 static const char	*output_postfix;
 
-/* Scan for more variables after one was found? */
-int			 morevars;
+size_t			 varsalloc, varscount;
+struct var		*vars;
 
 static const confget_backend	* const backends[] = {
 	&confget_http_get_backend,
@@ -72,9 +78,9 @@ static const confget_backend	* const backends[] = {
 };
 #define BACKEND_CNT	(sizeof(backends) / sizeof(backends[0]))
 
-static int		 find_backend(const char * const _name);
-static int		 init(int _argc, char * const *_argv);
-static void		 usage(int _error);
+static size_t		 find_backend(const char * const _name);
+static void		 init(int _argc, char * const *_argv);
+static void		 usage(bool _error);
 static void		 version(void);
 
 /* The currently used configuration backend (ini, http_get, etc.) */
@@ -84,20 +90,20 @@ const confget_backend	*backend;
 FILE			*conffile;
 
 /* Has the "-c" (check only) option been specified? */
-int			 cflag;
+bool			 cflag;
 /* In check-only mode, has the variable been found? */
-int			 cfound;
+bool			 cfound;
 /* Has the "-l" (list the section) option been specified? */
-static int		 lflag;
+static bool		 lflag;
 /* Has the "-L" (match varname against patterns) option been specified? */
-static int		 Lflag;
+static bool		 Lflag;
 /* Query for the names of the sections in the configuration file? */
-int			 qsections;
+bool			 qsections;
 /* Shell-quote the variable values? */
-static int		 Sflag;
+static bool		 Sflag;
 #ifdef HAVE_PCRE
 /* Treat patterns as regular expressions? */
-static int		 xflag;
+static bool		 xflag;
 
 static const pcre	*x_matchvalue, **x_margv;
 static const pcre_extra	*x_matchvalue_extra, **x_margv_extra;
@@ -115,16 +121,16 @@ static const pcre_extra	*x_matchvalue_extra, **x_margv_extra;
  *	error depending on the "error" flag.
  */
 void
-usage(int error)
+usage(const bool error)
 {
 	const char * const s1 = "Usage:\n"
-	    "confget [-cSx] [-N | -n] [-f filename] [-m pattern] [-P postfix]"
+	    "confget [-cOSx] [-N | -n] [-f filename] [-m pattern] [-P postfix]"
 	    " [-p prefix]\n"
 	    "        [-s section] [-t type] var...\n\n"
-	    "confget [-Sx] [-N | -n] [-f filename] [-m pattern] [-P postfix]"
+	    "confget [-OSx] [-N | -n] [-f filename] [-m pattern] [-P postfix]"
 	    " [-p prefix]\n"
 	    "        [-s section] [-t type] -L pattern...\n\n"
-	    "confget [-Sx] [-N | -n] [-f filename] [-m pattern] [-P postfix]"
+	    "confget [-OSx] [-N | -n] [-f filename] [-m pattern] [-P postfix]"
 	    " [-p prefix]\n"
 	    "        [-s section] [-t type] -l\n"
 	    "confget [-f filename] -q sections [-t type]\n\n"
@@ -138,7 +144,9 @@ usage(int error)
 	    "\t-l\tlist all variables in the specified section;\n"
 	    "\t-m\tonly display values that match the specified pattern;\n"
 	    "\t-N\talways display the variable name;\n"
-	    "\t-n\tnever display the variable name;\n";
+	    "\t-n\tnever display the variable name;\n"
+	    "\t-O\tallow variables in the specified section to override\n"
+	    "\t\tthose placed before any section definitions;\n";
 	const char * const s3 =
 	    "\t-P\tdisplay this string after the variable name;\n"
 	    "\t-p\tdisplay this string before the variable name;\n"
@@ -171,7 +179,7 @@ usage(int error)
 void
 version(void)
 {
-	printf("confget 1.05\n");
+	puts("confget 2.0.0");
 }
 
 /***
@@ -185,33 +193,23 @@ version(void)
  *	Sets section and varname on success.
  *	Writes to the standard error on error.
  */
-static int
-init(int argc, char * const *argv)
+static void
+init(const int argc, char * const * const argv)
 {
-	int bidx, ch, do_help, do_list, do_version, show_name, hide_name;
-	size_t i;
-#ifdef HAVE_PCRE
-	const char *pcre_err;
-	int pcre_ofs;
-	size_t t;
-#endif
-
 	conffile = NULL;
 
-	cflag = Lflag = lflag = Sflag = showvarname = manyvars = 0;
-	do_help = do_list = do_version = show_name = hide_name = 0;
+	cflag = Lflag = lflag = Sflag = showvarname = manyvars = override = false;
+	bool do_help = false, do_list = false, do_version = false,
+	     show_name = false, hide_name = false;
 	matchvalue = NULL;
 	output_prefix = output_postfix = "";
-	bidx = find_backend("ini");
-	if (bidx == -1) {
-		fputs("INTERNAL ERROR: no 'ini' backend\n", stderr);
-		return (-1);
-	}
-	while ((ch = getopt(argc, argv, "cf:hLlm:NnP:p:q:Ss:Tt:xV")) != EOF) {
+	size_t bidx = find_backend("ini");
+	int ch;
+	while ((ch = getopt(argc, argv, "cf:hLlm:NnOP:p:q:Ss:Tt:xV")) != EOF) {
 		switch (ch) {
 			case 'c':
-				cflag = 1;
-				manyvars = 0;
+				cflag = true;
+				manyvars = false;
 				break;
 
 			case 'f':
@@ -219,19 +217,19 @@ init(int argc, char * const *argv)
 				break;
 
 			case 'h':
-				do_help = 1;
+				do_help = true;
 				break;
 
 			case 'L':
-				Lflag = 1;
-				showvarname = 1;
-				manyvars = 1;
+				Lflag = true;
+				showvarname = true;
+				manyvars = true;
 				break;
 
 			case 'l':
-				lflag = 1;
-				showvarname = 1;
-				manyvars = 1;
+				lflag = true;
+				showvarname = true;
+				manyvars = true;
 				break;
 
 			case 'm':
@@ -239,11 +237,15 @@ init(int argc, char * const *argv)
 				break;
 
 			case 'N':
-				show_name = 1;
+				show_name = true;
 				break;
 
 			case 'n':
-				hide_name = 1;
+				hide_name = true;
+				break;
+
+			case 'O':
+				override = true;
 				break;
 
 			case 'P':
@@ -258,11 +260,11 @@ init(int argc, char * const *argv)
 				/* TODO: Implement better query parsing */
 				if (strcmp(optarg, "sections") != 0)
 					errx(1, "Only the 'sections' query is supported for the present");
-				qsections = 1;
+				qsections = true;
 				break;
 
 			case 'S':
-				Sflag = 1;
+				Sflag = true;
 				break;
 
 			case 's':
@@ -270,33 +272,28 @@ init(int argc, char * const *argv)
 				break;
 
 			case 'T':
-				do_list = 1;
+				do_list = true;
 				break;
 
 			case 't':
 				bidx = find_backend(optarg);
-				if (bidx == -1) {
-					fprintf(stderr, "No backend '%s'\n",
-					    optarg);
-					return (-1);
-				}
 				break;
 
 			case 'x':
 #ifdef HAVE_PCRE
-				xflag = 1;
+				xflag = true;
 				break;
 #else
 				errx(1, "No regular expression support in this confget build");
 #endif
 
 			case 'V':
-				do_version = 1;
+				do_version = true;
 				break;
 
 			case '?':
 			default:
-				usage(1);
+				usage(true);
 				break;
 		}
 	}
@@ -304,48 +301,43 @@ init(int argc, char * const *argv)
 	if (do_version)
 		version();
 	if (do_help)
-		usage(0);
+		usage(false);
 	if (do_list) {
 		printf("Supported backends:");
-		for (i = 0; i < BACKEND_CNT; i++)
+		for (size_t i = 0; i < BACKEND_CNT; i++)
 			printf(" %s", backends[i]->name);
 		puts("");
 	}
 	if (do_help || do_list || do_version)
 		exit(0);
 
-	argc -= optind;
-	argv += optind;
-	if (cflag && (lflag || Lflag)) {
-		warnx("The -c flag may not be used with -l or -L");
-		return (-1);
-	}
-	if (argc == 0 && !lflag && !qsections) {
-		warnx("No matching criteria specified");
-		return (-1);
-	} else if ((argc > 0) && lflag) {
-		warnx("Too many matching criteria specified");
-		return (-1);
-	}
-	if (argc > 1 && !lflag) {
-		showvarname = 1;
-		manyvars = 1;
+	margc = argc - optind;
+	margv = argv + optind;
+	if (cflag && (lflag || Lflag))
+		errx(1, "The -c flag may not be used with -l or -L");
+	if (margc == 0 && !lflag && !qsections)
+		errx(1, "No matching criteria specified");
+	if ((margc > 0) && lflag)
+		errx(1, "Too many matching criteria specified");
+	if (override && (section == NULL || section[0] == '\0'))
+		errx(1, "The -O flag only makes sense with a non-empty section name");
+	if (margc > 1 && !lflag) {
+		showvarname = true;
+		manyvars = true;
 	}
 	if (show_name) {
-		if (hide_name) {
-			warnx("The -N and -n flags may not be used together");
-			return (-1);
-		}
-		showvarname = 1;
+		if (hide_name)
+			errx(1, "The -N and -n flags may not be used together");
+		showvarname = true;
 	} else if (hide_name) {
-		showvarname = 0;
+		showvarname = false;
 	}
-	margc = argc;
-	margv = argv;
 
 #ifdef HAVE_PCRE
 	if (xflag) {
 		if (matchvalue) {
+			const char *pcre_err;
+			int pcre_ofs;
 			x_matchvalue = pcre_compile(matchvalue, 0, &pcre_err,
 			    &pcre_ofs, NULL);
 			if (pcre_err != NULL)
@@ -359,11 +351,13 @@ init(int argc, char * const *argv)
 		}
 		x_margv = malloc(margc * sizeof(*x_margv));
 		if (x_margv == NULL)
-			errx(1, "Could not allocate memory");
+			err(1, "Could not allocate memory");
 		x_margv_extra = malloc(margc * sizeof(*x_margv_extra));
 		if (x_margv_extra == NULL)
-			errx(1, "Could not allocate memory");
-		for (t = 0; t < (size_t)margc; t++) {
+			err(1, "Could not allocate memory");
+		for (size_t t = 0; t < (size_t)margc; t++) {
+			const char *pcre_err;
+			int pcre_ofs;
 			x_margv[t] = pcre_compile(margv[t], 0, &pcre_err,
 			    &pcre_ofs, NULL);
 			if (pcre_err != NULL)
@@ -380,7 +374,6 @@ init(int argc, char * const *argv)
 		errx(1, "The query for sections is only supported for the "
 		    "'ini' backend for the present");
 	backend = backends[bidx];
-	return (0);
 }
 
 /***
@@ -389,26 +382,26 @@ init(int argc, char * const *argv)
  * Inputs:
  *	name			- the name of the backend
  * Returns:
- *	The index in backends[] if found, -1 otherwise.
+ *	The index in backends[] if found; exits on error.
  * Modifies:
  *	Nothing.
  */
-static int
+static size_t
 find_backend(const char * const name)
 {
-	size_t i, len;
-	int tentative;
-	
-	len = strlen(name);
-	for (i = 0, tentative = -1; i < BACKEND_CNT; i++) {
-		if (strncmp(name, backends[i]->name, len))
+	const size_t len = strlen(name);
+	size_t tentative = BACKEND_CNT;
+	for (size_t i = 0; i < BACKEND_CNT; i++) {
+		if (strncmp(name, backends[i]->name, len) != 0)
 			continue;
 		if (backends[i]->name[len] == '\0')
 			return (i);
-		if (tentative >= 0)
-			return (-1);
+		if (tentative != BACKEND_CNT)
+			errx(1, "Ambiguous backend prefix '%s'", name);
 		tentative = i;
 	}
+	if (tentative == BACKEND_CNT)
+		errx(1, "Unknown backend '%s'", name);
 	return (tentative);
 }
 
@@ -418,20 +411,18 @@ find_backend(const char * const name)
  * Inputs:
  *	None.
  * Returns:
- *	0 on success, -1 on error.
+ *	Nothing; exits on error.
  * Modifies:
  *	Stores the opened file into fp.
  */
-static int
+static void
 openfile(void)
 {
-	if (backend->openfile == NULL) {
-		fprintf(stderr,
+	if (backend->openfile == NULL)
+		errx(2,
 		    "INTERNAL ERROR: backend '%s' does not define a openfile routine\n",
 		    backend->name);
-		return (-1);
-	}
-	return (backend->openfile());
+	backend->openfile();
 }
 
 /***
@@ -440,21 +431,19 @@ openfile(void)
  * Inputs:
  *	None.
  * Returns:
- *	0 on success, -1 on error.
+ *	Nothing; exits on error.
  * Modifies:
  *	May write to standard output if the variable has been found.
  */
-static int
+static void
 readfile(void)
 {
 
-	if (backend->readfile == NULL) {
-		fprintf(stderr,
+	if (backend->readfile == NULL)
+		errx(2,
 		    "INTERNAL ERROR: backend '%s' does not define a readfile routine\n",
 		    backend->name);
-		return (-1);
-	}
-	return (backend->readfile());
+	backend->readfile();
 }
 
 /***
@@ -463,21 +452,19 @@ readfile(void)
  * Inputs:
  *	None.
  * Returns:
- *	0 on success, -1 on error.
+ *	Nothing; exits on error.
  * Modifies:
  *	Closes the file pointed to by fp.
  */
-static int
+static void
 closefile(void)
 {
 
-	if (backend->closefile == NULL) {
-		fprintf(stderr,
+	if (backend->closefile == NULL)
+		errx(2,
 		    "INTERNAL ERROR: backend '%s' does not define a closefile routine\n",
 		    backend->name);
-		return (-1);
-	}
-	return (backend->closefile());
+	backend->closefile();
 }
 
 /***
@@ -488,79 +475,111 @@ closefile(void)
  *	name		- the variable name
  *	value		- the variable value
  * Returns:
- *	0 on success, -1 on failure.
+ *	Nothing; exits on error.
  * Modifies:
  *	In check-only mode, sets cfound.
  *	In normal mode, writes to standard output.
  */
-int
-foundvar(const char * const sec __unused, const char * const name,
+void
+foundvar(const char * const sec, const char * const name,
 	const char * const value)
 {
-	int i;
-#ifdef HAVE_PCRE
-	int r;
-#endif
-
 	if (!lflag) {
-		for (i = 0; i < margc; i++)
+		bool found = false;
+		for (int i = 0; i < margc; i++)
 			if (Lflag) {
 #ifdef HAVE_PCRE
 				if (xflag) {
-					r = pcre_exec(x_margv[i],
+					const int r = pcre_exec(x_margv[i],
 					    x_margv_extra[i], name,
 					    strlen(name), 0, 0, NULL, 0);
-					if (r == 0)
+					if (r == 0) {
+						found = true;
 						break;
+					}
 					if (r != PCRE_ERROR_NOMATCH)
 						errx(1, "Could not match '%s' against the '%s' pattern", name, margv[i]);
 				} else
 #endif
-				if (fnmatch(margv[i], name, 0) == 0)
+				if (fnmatch(margv[i], name, 0) == 0) {
+					found = true;
 					break;
+				}
 			} else {
-				if (!strcmp(name, margv[i]))
+				if (strcmp(name, margv[i]) == 0) {
+					found = true;
 					break;
+				}
 			}
-		if (i == margc) {
-			morevars = 1;
-			return (0);
-		}
+		if (!found)
+			return;
 	}
 
+	bool display = true;
 	if (matchvalue != NULL) {
-		i = 1;
 #ifdef HAVE_PCRE
 		if (xflag) {
-			r = pcre_exec(x_matchvalue, x_matchvalue_extra, value,
+			const int r = pcre_exec(x_matchvalue, x_matchvalue_extra, value,
 			    strlen(value), 0, 0, NULL, 0);
 			if (r == PCRE_ERROR_NOMATCH)
-				i = 0;
+				display = false;
 			else if (r != 0)
-				errx(1, "Could not match '%s' against the '%s' pattern", value, margv[i]);
+				errx(1, "Could not match '%s' against the '%s' pattern", value, matchvalue);
 		} else
 #endif
-		if (fnmatch(matchvalue, value, 0) == FNM_NOMATCH) {
-			i = 0;
-		}
-		if (i == 0) {
-			morevars = 1;
-			return (0);
-		}
+		if (fnmatch(matchvalue, value, 0) == FNM_NOMATCH)
+			display = false;
+		if (!display && !override)
+			return;
 	}
-	if (cflag) {
-		cfound = 1;
-		morevars = 0;
-	} else {
+
+	for (size_t i = 0; i < varscount; i++)
+	{
+		struct var * const var = &vars[i];
+		if (strcmp(var->name, name) != 0)
+			continue;
+		free(var->value);
+		var->value = strdup(value);
+		free(var->section);
+		var->section = sec == NULL? NULL: strdup(sec);
+		var->display = display;
+		return;
+	}
+	if (varscount == varsalloc) {
+		varsalloc += 16;
+		vars = realloc(vars, varsalloc * sizeof(*vars));
+		if (vars == NULL)
+			err(1, "Could not allocate memory for the variables' values");
+	}
+	vars[varscount].name = strdup(name);
+	vars[varscount].value = strdup(value);
+	vars[varscount].section = sec == NULL? NULL: strdup(sec);
+	vars[varscount].display = display;
+	varscount++;
+}
+
+static void
+displayvars(void)
+{
+	cfound = false;
+	for (size_t i = 0; i < varscount; i++)
+	{
+		if (!vars[i].display)
+			continue;
+		if (cflag) {
+			cfound = true;
+			return;
+		}
+		const char * const name = vars[i].name;
+		const char * const value = vars[i].value;
+
 		if (showvarname)
 			printf("%s%s%s=", output_prefix, name, output_postfix);
 		if (!Sflag) {
 			printf("%s\n", value);
 		} else {
-			const char *p;
-
 			printf("'");
-			p = value;
+			const char *p = value;
 			while (*p) {
 				while (*p && *p != '\'')
 					putchar(*p++);
@@ -573,9 +592,7 @@ foundvar(const char * const sec __unused, const char * const name,
 			}
 			printf("'\n");
 		}
-		morevars = manyvars;
 	}
-	return (0);
 }
 
 /***
@@ -584,70 +601,52 @@ foundvar(const char * const sec __unused, const char * const name,
  * Inputs:
  *	sec		- the name of the new section
  * Returns:
- *	0 on success, -1 on failure.
+ *	Nothing; exits on error.
  * Modifies:
  *	In "-q sections" mode, writes to standard output.
  */
-int
+void
 foundsection(const char * const _sec)
 {
+	if (!qsections)
+		return;
+
 	static size_t alloc = 0, count = 0;
 	static char **names = NULL;
-	size_t i;
-
-	if (!qsections)
-		return (0);
-
-	for (i = 0; i < count; i++)
+	for (size_t i = 0; i < count; i++)
 		if (strcmp(_sec, names[i]) == 0)
-			return (0);
+			return;
 	/* A new section! */
 	if (alloc == count) {
-		size_t nalloc;
-		char **newnames;
-		
-		nalloc = 2 * alloc + 1;
-		newnames = realloc(names, nalloc * sizeof(names[0]));
-		if (newnames == NULL) {
-			warnx("Out of memory for the section names list");
-			return (-1);
-		}
+		const size_t nalloc = 2 * alloc + 1;
+		char ** const newnames = realloc(names, nalloc * sizeof(names[0]));
+		if (newnames == NULL)
+			err(1, "Out of memory for the section names list");
 		names = newnames;
 		alloc = nalloc;
 	}
 	names[count] = strdup(_sec);
-	if (names[count] == NULL) {
-		warnx("Out of memory for the new section name");
-		return (-1);
-	}
+	if (names[count] == NULL)
+		err(1, "Out of memory for the new section name");
 	count++;
 
 	/* And finally output it :) */
 	puts(_sec);
-	return (0);
 }
 
 /***
  * Main routine
  */
 int
-main(int argc, char * const *argv)
+main(const int argc, char * const * const argv)
 {
 
-	if (init(argc, argv) == -1)
-		return (1);
+	init(argc, argv);
 
-	cfound = 0;
-	if (openfile() == -1)
-		return (1);
-	if (readfile() == -1) {
-		closefile();
-		return (1);
-	}
-	if (closefile() == -1)
-		return (1);
+	openfile();
+	readfile();
+	closefile();
 
-	if (cflag)
-		return (!cfound);
-	return (0);
+	displayvars();
+	return cflag? !cfound: 0;
 }
